# Standard lib imports
import socket

# Local imports
from qwilfish.base_runner import BaseRunner

class LinuxSocketRunner(BaseRunner):

    def __init__(self, interface="lo"):
        self.sock = socket.socket(socket.AF_PACKET, socket.SOCK_RAW)
        self.sock.bind((interface, 0))

    def run(self, input):
        data = []
        for byte in range(0, len(input), 8):
            data.append(int(input[byte:byte+8], 2))

        frame = bytes(data)

        self.sock.send(frame)



# Standard lib imports
import random
import re


# Local imports
from qwilfish.constants import DEFAULT_START_SYMBOL
from qwilfish.grammar import opts
from qwilfish.derivation_tree import all_terminals

# TODO fuzz by changing endianness of fields and/or bytes
# TODO fuzz by inserting wrong TLVs

HEX_MAP = {"0": "0000", "1": "0001", "2": "0010", "3": "0011",
           "4": "0100", "5": "0101", "6": "0110", "7": "0111",
           "8": "1000", "9": "1001", "A": "1010", "B": "1011",
           "C": "1100", "D": "1101", "E": "1110", "F": "1111"}

def srange(characters):
    return [c for c in characters]

def to_binstr(subtree, *args):
    symbol, children = subtree

    if children == []:
        if symbol == "x":
            return ("", []) # Turn single x:es into null expansions to remove
        else:
            new_symbol = \
                re.sub("x[0-9a-fA-F]{2}",
                       lambda hex: HEX_MAP[hex.group(0)[1:].upper()[0]] +
                                   HEX_MAP[hex.group(0)[1:].upper()[1]],
                       symbol)
            return (new_symbol, [])
    else:
        if symbol == "<hex>":
            child_symbol, _ = children[0]
            return (symbol, [(HEX_MAP[child_symbol], [])])

    return (symbol, [to_binstr(c, "") for c in children if children])

def fix_length(subtree, *args): # TODO make cleaner, avoid all_terminals
    symbol, children = subtree
    n_bits = len(all_terminals(subtree)) - 16 # 16 bits for type+length

    assert n_bits % 8 == 0

    length = int(n_bits/8)
    bin_length = ["1" if length & (1 << i) else "0" for i in range(8, -1, -1)]

    for i, c in enumerate(children):
        if c[0] == "<tlv-len>":
            children[i] = (children[i][0],
                           [("".join(bin_length), [])])

    return (symbol, children)

def gen_random_data():
    max_size = 255
    length = random.randint(1,255)*8
    bin_data = [str(random.randint(0,1)) for i in range(0, length)]

    return "".join(bin_data)

ETHERNET_FRAME_GRAMMAR = {
    DEFAULT_START_SYMBOL  : [("<ethernet-frame>", opts(post=to_binstr))],
    "<ethernet-frame>"    : ["<addr><vlan-tags><type-payload>"],
    "<addr>"              : ["<dst><src>"],
    "<dst>"               : [("<byte><byte><byte><byte><byte><byte>",
                              opts(prob=0.1)),
                             "<mef-multicast>"
                            ],
    "<mef-multicast>"     : ["x01x80xC2x00x00x00", "x01x80xC2x00x00x03",
                             "x01x80xC2x00x00x0E"
                            ],
    "<src>"               : ["<byte><byte><byte><byte><byte><byte>"],
    "<vlan-tags>"         : ["", "<q-tag><vlan-tags>", "<q-tag>"],
    "<q-tag>"             : ["<tpid><pcp><dei><vlan>"],
    "<tpid>"              : ["x81x00", "x88xA8"],
    "<pcp>"               : ["<bit><bit><bit>"],
    "<dei>"               : ["<bit>"],
    "<vlan>"              : ["<byte><bit><bit><bit><bit>"],
    "<type-payload>"      : ["<lldp-ethertype><lldp-payload>"],
    "<byte>"              : ["x<hex><hex>"],
    "<hex>"               : srange("0123456789ABCDEF"),
    "<bit>"               : ["0", "1"],
    "<random-data>"       : [("", opts(pre=gen_random_data))],
    "<lldp-ethertype>"    : ["x88xCC"],
    "<lldp-payload>"      : ["<lldp-tlv-chassiid><lldp-tlv-portid>" \
                             "<lldp-tlv-ttl><lldp-opt-tlvs><lldp-tlv-end>"],
    "<lldp-opt-tlvs>"     : ["", "<lldp-opt-tlv>",
                             "<lldp-opt-tlv><lldp-opt-tlvs>"
                            ],
    "<lldp-opt-tlv>"      : ["<lldp-tlv-portdesc>", "<lldp-tlv-sysname>",
                             "<lldp-tlv-sysdesc>", "<lldp-tlv-syscap>",
                             "<lldp-tlv-mgmtaddr>", "<lldp-tlv-custom>"
                            ],
    "<tlv-len>"           : ["<bit><byte>"],
    "<lldp-tlv-end>"      : ["x00x00"],
    "<lldp-tlv-chassiid>" : [("0000001<tlv-len><chassiid-subtype>"
                             "<chassiid-data>", opts(post=fix_length))],
    "<chassiid-subtype>"  : ["00000001", "00000010", "00000011", "00000100",
                             "00000101", "00000110", "00000111"],
    "<chassiid-data>"     : ["<random-data>"],
    "<lldp-tlv-portid>"   : [("0000010<tlv-len><portid-subtype><portid-data>",
                              opts(post=fix_length))],
    "<portid-subtype>"    : ["00000001", "00000010", "00000011", "00000100",
                             "00000101", "00000110", "00000111"],
    "<portid-data>"       : ["<random-data>"],
    "<lldp-tlv-ttl>"      : [("0000011<tlv-len><byte><byte>",
                              opts(post=fix_length))],
    #"<lldp-tlv-portdesc>" : [("0000100<tlv-len><chassiid-subtype><chassiid-data>", opts(post=fix_length))],
    #"<lldp-tlv-sysname>"  : [("0000101<tlv-len><chassiid-subtype><chassiid-data>", opts(post=fix_length))],
    #"<lldp-tlv-sysdesc>"  : [("0000110<tlv-len><chassiid-subtype><chassiid-data>", opts(post=fix_length))],
    #"<lldp-tlv-syscap>"   : [("0000111<tlv-len>", opts(post=fix_length))],
    #"<lldp-tlv-mgmtaddr>" : [("0001000<tlv-len>", opts(post=fix_length))],
    #"<lldp-tlv-custom>"   : [("1111111<tlv-len>", opts(post=fix_length))]
    "<lldp-tlv-portdesc>" : [""],
    "<lldp-tlv-sysname>"  : [""],
    "<lldp-tlv-sysdesc>"  : [""],
    "<lldp-tlv-syscap>"   : [""],
    "<lldp-tlv-mgmtaddr>" : [""],
    "<lldp-tlv-custom>"   : [""]
}



'''
Unit tests for the main functionality of the  qwilfuzzer module.
'''

# Standard library imports
import unittest

# Local imports
import qwilfish.qwilfuzzer as qqf

class TestQwilfuzzerDummy(unittest.TestCase):
    '''
    Dummy test class
    '''

    def test_dummy(self):
        '''
        Dummy test
        '''
        self.assertTrue(True)

if __name__ == "__main__":
    unittest.main()



# Standard library imports
import argparse

# Local imports
from qwilfish.qwilfuzzer import QwilFuzzer
from qwilfish.constants import SIMPLE_GRAMMAR_EXAMPLE
from qwilfish.ethernet_frame import ETHERNET_FRAME_GRAMMAR
from qwilfish.linux_socket_runner import LinuxSocketRunner

def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("-l", "--log",
                        dest="log",
                        help="Turn on logging",
                        action="store_true")

    parser.add_argument("-i", "--interface",
                        dest="interface",
                        help="Network interface to transmit fuzzed packets on",
                        default="lo")

    parser.add_argument("-c", "--count",
                        dest="count",
                        help="Number of fuzzed packets to transmit",
                        type=int,
                        default=1)


    args = parser.parse_args()

    grammar = ETHERNET_FRAME_GRAMMAR
    runner = LinuxSocketRunner(interface=args.interface)
    qf = QwilFuzzer(grammar, log=args.log)
    for i in range(0, args.count):
        qf.run(runner=runner)



'''
Unit tests for the main functionality of the display_tree module.
'''

# Standard library imports
import unittest

# Local imports
import qwilfish.derivation_tree as qdt

class TestDerivationTreeAllTerminals(unittest.TestCase):
    '''
    Test the all_terminals function in the display_tree module.
    '''

    def test_all_terminals_single_nonterminal(self):
        '''
        Test with a single nonterminal symbol.
        '''
        symbol = "<start>"
        tree = (symbol, None)
        self.assertEqual(qdt.all_terminals(tree), symbol)


    def test_all_terminals_single_terminal(self):
        '''
        Test with a single terminal symbol.
        '''
        symbol = "terminal"
        tree = (symbol, [])
        self.assertEqual(qdt.all_terminals(tree), symbol)

    def test_all_terminals_epsilon(self):
        '''
        Test with an epsilon expansion.
        '''
        symbol = ""
        tree = (symbol, [])
        self.assertEqual(qdt.all_terminals(tree), symbol)

    def test_all_terminals_two_level_nonterminal(self):
        '''
        Test with a two-level tree where the leaves are nonterminals.
        '''
        symbol_2a = "<symbol-2a>"
        symbol_2b = "<symbol-2b>"
        tree = ("<level-1>", [(symbol_2a, None), (symbol_2b, None)])
        expected = symbol_2a + symbol_2b
        self.assertEqual(qdt.all_terminals(tree), expected)

    def test_all_terminals_two_level_terminal(self):
        '''
        Test with a two-level tree where the leaves are terminals.
        '''
        symbol_2a = "terminal-2a"
        symbol_2b = "terminal-2b"
        tree = ("<level-1>", [(symbol_2a, []), (symbol_2b, [])])
        expected = symbol_2a + symbol_2b
        self.assertEqual(qdt.all_terminals(tree), expected)

    def test_all_terminals_two_level_mix(self):
        '''
        Test with a two-level tree with both terminals and nonterminals.
        '''
        symbol_2a = "terminal-2a"
        symbol_2b = "<nonterminal-2b>"
        symbol_2c = "terminal-2c"
        tree = ("<level-1>",
                [(symbol_2a, []), (symbol_2b, None), (symbol_2c, [])])
        expected = symbol_2a + symbol_2b + symbol_2c
        self.assertEqual(qdt.all_terminals(tree), expected)

class TestDerivationTreeExpansionToChildren(unittest.TestCase):
    '''
    Test the expansion_to_children function in the display_tree module.
    '''

    def test_expansion_to_children_single_nonterminal(self):
        '''
        Test with a single nonterminal symbol.
        '''
        self.assertEqual(qdt.expansion_to_children("<a>"),
                         [("<a>", None)])

    def test_expansion_to_children_single_terminal(self):
        '''
        Test with a single terminal symbol.
        '''
        self.assertEqual(qdt.expansion_to_children("a"),
                         [("a", [])])

    def test_expansion_to_children_list(self):
        '''
        Test with a wrongly typed expansion 2-tuple, first ele must be a str
        '''
        self.assertRaises(TypeError, qdt.expansion_to_children, ([7], []))

    def test_expansion_to_children_mix(self):
        '''
        Test with a mix of terminals and nonterminals
        '''
        self.assertEqual(qdt.expansion_to_children("<a>b<c>12"),
                         [("<a>", None), ("b", []), ("<c>", None), ("12", [])])





if __name__ == "__main__":
    unittest.main()



'''
Unit tests for the main functionality of the grammar module.
'''

# Standard library imports
import unittest

# Local imports
import qwilfish.grammar as qg
from qwilfish.constants import DEFAULT_START_SYMBOL

# Grammars used for some test cases
TRIVIAL_GRAMMAR = {
    DEFAULT_START_SYMBOL : ["<phrase>"],
    "<phrase>"           : ["<hello><world>"],
    "<hello>"            : ["Hello, "],
    "<world>"            : ["world!"]
}

NONE_GRAMMAR = None

EMPTY_GRAMMAR = {}

WRONG_TYPE_GRAMMAR = ("<abc>", "def")

BAD_SYMBOL_GRAMMAR = {
    DEFAULT_START_SYMBOL : ["<illegal-symbol>"],
    "<illegal-symbol>"   : ["<<> >"],
    "<<> >"              : ["Symbol names may not contain >, < or space!"]
}

ETERNAL_GRAMMAR = {
    DEFAULT_START_SYMBOL : ["<symbol-1>"],
    "<symbol-1>"         : ["<symbol-2>"],
    "<symbol-2>"         : ["<symbol-1>"]
}

UNDEFINED_SYMBOL_GRAMMAR = {
    DEFAULT_START_SYMBOL : ["<defined-sym>"],
    "<defined-sym>"      : ["terminal", "<undefined-sym>"]
}

UNUSED_SYMBOL_GRAMMAR = {
    DEFAULT_START_SYMBOL : ["<used-sym>"],
    "<used-sym>"         : ["terminal", "another terminal"],
    "<unused-sym>"       : ["unused terminal"]
}

class TestGrammarValidate(unittest.TestCase):
    '''
    Test the validate function in the grammar module.
    '''

    def test_validate_trivial(self):
        '''
        A trivial grammar that should be valid.
        '''
        self.assertTrue(qg.validate(TRIVIAL_GRAMMAR))

    def test_validate_none(self):
        '''
        None-type objects are not valid grammars.
        '''
        self.assertFalse(qg.validate(NONE_GRAMMAR))

    def test_validate_wrong_type(self):
        '''
        Non-dict objects (in this case a tuple) are not valid grammars.
        '''
        self.assertFalse(qg.validate(WRONG_TYPE_GRAMMAR))

    def test_validate_empty(self):
        '''
        Empty dicts are not valid grammars.
        '''
        self.assertFalse(qg.validate(EMPTY_GRAMMAR))

    def test_validate_bad_symbol(self):
        '''
        Nonterminals containing >< or space make the grammar invalid.
        '''
        self.assertFalse(qg.validate(BAD_SYMBOL_GRAMMAR))

    def test_validate_eternal(self):
        '''
        Grammars that can never be completely expanded are valid (for now).
        '''
        self.assertTrue(qg.validate(ETERNAL_GRAMMAR))

    def test_validate_undefined_symbol(self):
        '''
        All nonterminals in the expansions/values must also be used as a key.
        '''
        self.assertFalse(qg.validate(UNDEFINED_SYMBOL_GRAMMAR))

    def test_validate_unused_symbol(self):
        '''
        All keys/symbols must also be used in at least one expansion/value.
        '''
        self.assertFalse(qg.validate(UNUSED_SYMBOL_GRAMMAR))



class TestGrammarNonterminals(unittest.TestCase):
    '''
    Test the nonterminals function in the grammar module.
    '''

    def test_nonterminals_basic(self):
        '''
        Find the two nonterminals in an expansion string with nothing else.
        '''
        self.assertEqual(qg.nonterminals("<a><b>"), ["<a>", "<b>"])

    def test_nonterminals_mixed(self):
        '''
        Find the two nonterminals in an expansion string with a mix of symbols.
        '''
        self.assertEqual(qg.nonterminals("<a>b<c>"), ["<a>", "<c>"])

    def test_nonterminals_bad_symbol(self):
        '''
        Don't find nonterminals with illegal names.
        '''
        self.assertEqual(qg.nonterminals("< a>"), [])

    def test_nonterminals_midst(self):
        '''
        Find a nonterminal in the midst of a string of mostly terminals.
        '''
        self.assertEqual(qg.nonterminals("abc<d>efg"), ["<d>"])

    def test_nonterminals_terminal_only(self):
        '''
        Don't find anything in a string of only terminals.
        '''
        self.assertEqual(qg.nonterminals("a b"), [])

    def test_nonterminals_annotated(self):
        '''
        Find a single annotated nonterminal symbol.
        '''
        self.assertEqual(qg.nonterminals(("<a>", {"opt": "val"})), ["<a>"])


class TestGrammarIsNonterminal(unittest.TestCase):
    '''
    Test the is_nonterminal function in the grammar module.
    '''

    def test_is_nonterminal_basic(self):
        '''
        Nonterminal symbol.
        '''
        self.assertTrue(qg.is_nonterminal("<symbol>"))

    def test_is_nonterminal_complicated_name(self):
        '''
        Nonterminal symbol with a slightly complicated name.
        '''
        self.assertTrue(qg.is_nonterminal("<complicated-symbol-123>"))

    def test_is_nonterminal_not(self):
        '''
        Terminal symbol.
        '''
        self.assertFalse(qg.is_nonterminal("terminal"))

    def test_is_nonterminal_bad_symbol_name(self):
        '''
        Nonterminal symbol with illegal name.
        '''
        self.assertFalse(qg.is_nonterminal("<<> <<> <>"))

    def test_is_nonterminal_mixed_terminal_nonterminal(self):
        '''
        Nonterminals must begin with <
        '''
        self.assertFalse(qg.is_nonterminal("a<b>"))

    def test_is_nonterminal_mixed_nonterminal_terminal(self):
        '''
        Nonterminals followed by terminals will be found
        '''
        self.assertTrue(qg.is_nonterminal("<a>b"))

    def test_is_nonterminal_missing_closing(self):
        '''
        Nonterminals must end with >
        '''
        self.assertFalse(qg.is_nonterminal("<ab"))

    def test_is_nonterminal_missing_opening(self):
        '''
        Nonterminals must begin with <
        '''
        self.assertFalse(qg.is_nonterminal("ab>"))

class TestGrammarExpansionString(unittest.TestCase):
    '''
    Test the expansion_string function in the grammar module.
    '''

    def test_is_expansion_string_empty_opts(self):
        '''
        Return symbol string in an expansion tuple with an empty dict
        '''
        symbol = "<symbol>"
        self.assertEqual(qg.expansion_string((symbol, {})), symbol)

    def test_is_expansion_string_no_opts(self):
        '''
        Return symbol string in an expansion that is not a tuple
        '''
        symbol = "<symbol>"
        self.assertEqual(qg.expansion_string(symbol), symbol)

    def test_is_expansion_string_wrong_type_int(self):
        '''
        Make sure error is raised if an int is given as the argument
        '''
        self.assertRaises(TypeError, qg.expansion_string, 7)

    def test_is_expansion_string_wrong_type_3tuple(self):
        '''
        Make sure error is raised if a 3-tuple is given as the argument
        '''
        self.assertRaises(TypeError, qg.expansion_string, ("a", {}, "b"))

    def test_is_expansion_string_wrong_type_2tuple(self):
        '''
        Make sure error is raised for 2-tuple with the wrong type in first pos
        '''
        self.assertRaises(TypeError, qg.expansion_string, (7, {}))

@unittest.skip("Expansion opts not supported yet, skipping")
class TestGrammarExpansionOpts(unittest.TestCase):
    '''
    Test the expansion_opts function in the grammar module.
    '''

    def test_is_expansion_opts_basic(self):
        '''
        An expansion that is just a string should yield an empty dict
        '''
        self.assertEqual(qg.expansion_opts("<symbol>"), {})

if __name__ == "__main__":
    unittest.main()



class BaseRunner:
    '''Base class for creating test cases from the fuzzed input'''

    RET_PASS = "PASS"
    RET_FAIL = "FAIL"
    RET_URES = "UNRESOLVED"

    def __init__(self):
        '''Initialize'''
        pass

    def run(self, input):
        '''Deliver the input to the SUT and return result and whatever
        quantities are being monitored'''
        print(input)
        return (input, BaseRunner.RET_URES)



# Standard lib imports
import random

# Local imports
from qwilfish.constants import DEFAULT_START_SYMBOL
from qwilfish.constants import DEFAULT_MIN_NONTERMINALS
from qwilfish.constants import DEFAULT_MAX_NONTERMINALS
from qwilfish.base_runner import BaseRunner
from qwilfish.grammar import GrammarError
from qwilfish.grammar import validate
from qwilfish.grammar import nonterminals
from qwilfish.grammar import is_nonterminal
from qwilfish.grammar import expansion_opt
from qwilfish.grammar import expansion_string
from qwilfish.derivation_tree import DerivationTreeError
from qwilfish.derivation_tree import expansion_to_children
from qwilfish.derivation_tree import all_terminals

class QwilFuzzer():

    OPT_PRE = "pre"
    OPT_POST = "post"
    OPT_PROB = "prob"
    DEFAULT_OPTS = [OPT_PRE, OPT_POST, OPT_PROB]

    def __init__(self,
                 grammar,
                 start_symbol=DEFAULT_START_SYMBOL,
                 min_nonterminals=DEFAULT_MIN_NONTERMINALS,
                 max_nonterminals=DEFAULT_MAX_NONTERMINALS,
                 supported_opts=DEFAULT_OPTS,
                 log=False):
        self.grammar = grammar
        self.start_symbol = start_symbol
        self.min_nonterminals = min_nonterminals
        self.max_nonterminals = max_nonterminals

        self.supported_opts = set()
        if supported_opts:
            for i in range(0, len(supported_opts)):
                self.supported_opts |= {supported_opts[i]}

        self.log=log
        self.check_grammar()

    def fuzz(self):
        self.derivation_tree = self.fuzz_tree()
        return all_terminals(self.derivation_tree)

    def run(self, runner=BaseRunner()):
        return runner.run(self.fuzz())

    def check_grammar(self):
        if self.start_symbol not in self.grammar:
            raise GrammarError("Start symbol" + self.start_symbol +
                               "not found in supplied grammar")
        if not validate(self.grammar, self.start_symbol,
                        self.supported_opts):
            raise GrammarError("Invalid grammar supplied")
        for nonterminal in self.grammar:
            expansions = self.grammar[nonterminal]
            _ = self.expansion_probabilities(expansions, nonterminal)

        return True

###############################################################################
############################################################### Private methods
###############################################################################

    def expand_node(self, node):
        return self.expand_node_randomly(node)

    def init_tree(self):
        return (self.start_symbol, None)

    def expansion_to_children(self, expansion):
        return expansion_to_children(expansion)

    def expand_node_min_cost(self, node):
        if self.log:
            print("Expanding", all_terminals(node), "at minimum cost")

        return self.expand_node_by_cost(node, min)

    def expand_node_max_cost(self, node):
        if self.log:
            print("Expanding", all_terminals(node), "at maximum cost")

        return self.expand_node_by_cost(node, max)

    def expand_node_randomly(self, node):
        (symbol, children) = node
        if not children is None:
            raise DerivationTreeError("Unexpandable node, children: " +
                                      children)

        if self.log:
            print("Expanding", all_terminals(node), "randomly")

        expansions = self.grammar[symbol]
        children_alternatives = \
            [self.expansion_to_children(expansion) for expansion in expansions]

        index = self.choose_node_expansion(node, children_alternatives)
        chosen_children = children_alternatives[index]

        chosen_children = self.process_chosen_children(chosen_children,
                                                       expansions[index])

        return (symbol, chosen_children)

    def opt_pre_func(self, expansion):
        return expansion_opt(expansion, QwilFuzzer.OPT_PRE)

    def opt_post_func(self, expansion):
        return expansion_opt(expansion, QwilFuzzer.OPT_POST)

    def opt_prob(self, expansion):
        return expansion_opt(expansion, QwilFuzzer.OPT_PROB)

    def process_chosen_children(self, children, expansion):
        function = self.opt_pre_func(expansion)
        if function is None:
            return children

        if not callable(function):
            raise GrammarError("Opt function 'pre' not callable")

        result = function()

        if self.log:
            print(repr(function) + "()", "=", repr(result))

        return self.apply_result(result, children)

    def apply_result(self, result, children):
        if isinstance(result, str):
            children = [(result, [])]
        elif isinstance(result, list):
            symbol_indexes = [i for i, c in enumerate(children)
                              if is_nonterminal(c[0])] # Always a tuple here

            for index, value in enumerate(result):
                if value is not None:
                    child_index = symbol_indexes[index]
                    if not isinstance(value, str):
                        value = repr(value)
                    if self.log:
                        print("Replacing", all_terminals(
                            children[child_index]), "by", value)

                    child_symbol, _ = children[child_index]
                    children[child_index] = (child_symbol, [(value, [])])
        elif isinstance(result, tuple):
            _, children = result
        elif result is None:
            pass
        elif isinstance(result, bool):
            pass
        else:
            if self.log:
                print("Replacing", "".join(
                    [all_terminals(c) for c in children]), "by", result)

            children = [(repr(result), [])]

        return children

    def choose_node_expansion(self, node, children_alternatives):
        (symbol, tree) = node
        expansions = self.grammar[symbol]
        probabilities = self.expansion_probabilities(expansions)

        weights = []
        for children in children_alternatives:
            expansion = all_terminals((symbol, children))
            children_weight = probabilities[expansion]
            if self.log:
                print(repr(expansion), "p =", children_weight)
            weights.append(children_weight)

        if sum(weights) == 0:
            return random.choices(range(len(children_alternatives)))[0]
        else:
            return random.choices(range(len(children_alternatives)),
                                  weights=weights)[0]

    def possible_expansions(self, node):
        (symbol, children) = node
        if children is None:
            return 1

        return sum(self.possible_expansions(c) for c in children)

    def any_possible_expansions(self, node):
        (symbol, children) = node
        if children is None:
            return True

        return any(self.any_possible_expansions(c) for c in children)

    def choose_tree_expansion(self, tree, children):
        return random.randrange(0, len(children))

    def expand_tree_once(self, tree):
        (symbol, children) = tree
        if children is None:
            return self.expand_node(tree)

        expandable_children = [
            c for c in children if self.any_possible_expansions(c)]

        index_map = [i for (i, c) in enumerate(children)
                     if c in expandable_children]

        child_to_be_expanded = \
            self.choose_tree_expansion(tree, expandable_children)

        children[index_map[child_to_be_expanded]] = \
            self.expand_tree_once(expandable_children[child_to_be_expanded])

        return tree

    def fuzz_tree(self):
        while True:
            tree = self.init_tree()
            tree = self.expand_tree(tree)
            symbol, children = tree
            result, new_children = self.run_post_functions(tree)
            if not isinstance(result, bool) or result:
                return (symbol, new_children)
            self.restart_expansion()

    def restart_expansion(self):
        pass

    def run_post_functions(self, tree, depth=float("inf")):
        symbol, children = tree

        if children == []:
            return True, children # Terminal symbol

        try:
            expansion = self.find_expansion(tree)
        except KeyError:
            return True, children # Expansion not found, ignore

        result = True
        function = self.opt_post_func(expansion)
        if function is not None:
            result = self.eval_function(tree, function)
            if isinstance(result, bool) and not result:
                if self.log:
                    print(all_terminals(tree),
                        "did not satisfy", symbol, "constraint")
                return False, children

            children = self.apply_result(result, children)

        if depth > 0:
            for c in children:
                result, _ = self.run_post_functions(c, depth-1)
                if isinstance(result, bool) and not result:
                    return False, children

        return result, children

    def find_expansion(self, tree):
        symbol, children = tree

        applied_expansion = \
            "".join([child_symbol for child_symbol, _ in children])

        # TODO Some intelligent mechanism to find expansions with
        # hexstring literals that have been replaced by binstrings.
        # Currently, it is not possible to have a 'post' opt if
        # hex literals are also being used in the same expression
        for expansion in self.grammar[symbol]:
            if expansion_string(expansion) == applied_expansion:
                return expansion

        raise KeyError(
            symbol +
            ": did not find expansion " +
            repr(applied_expansion))

    def eval_function(self, tree, function):
        symbol, children = tree

        if not callable(function):
            raise GrammarError("Opt function 'post' not callable")

        args = []
        for (symbol, exp) in children:
            if exp != [] and exp is not None:
                symbol_value = all_terminals((symbol, exp))
                args.append(symbol_value)

        result = function(tree, *args)
        if self.log:
            print(repr(function) + repr(tuple(args)), "=", repr(result))

        return result

    def expand_tree(self, tree):
        tree = self.expand_tree_with_strategy(
            tree, self.expand_node_max_cost, self.min_nonterminals)
        tree = self.expand_tree_with_strategy(
            tree, self.expand_node_randomly, self.max_nonterminals)
        tree = self.expand_tree_with_strategy(
            tree, self.expand_node_min_cost)

        if not self.possible_expansions(tree) == 0:
            raise DerivationTreeError("Unable to expand tree to completion")

        return tree

    def expand_tree_with_strategy(self,
                                  tree,
                                  expand_node_method,
                                  limit = None):
        self.expand_node = expand_node_method
        while ((limit is None
                or self.possible_expansions(tree) < limit)
                and self.any_possible_expansions(tree)):
            tree = self.expand_tree_once(tree)
            self.log_tree(tree)

        return tree

    def expand_node_by_cost(self, node, choose = min):
        (symbol, children) = node

        if not children is None:
            raise DerivationTreeError("Unexpandable node, children: " +
                                      children)

        expansions = self.grammar[symbol]

        children_alternatives_with_cost = [
            (self.expansion_to_children(expansion),
             self.expansion_cost(expansion, {symbol}),
             expansion)
            for expansion in expansions]

        costs = [cost for (child, cost, expansion)
                 in children_alternatives_with_cost]

        chosen_cost = choose(costs)
        children_with_chosen_cost = [child for (child, child_cost, _)
                                     in children_alternatives_with_cost
                                     if child_cost == chosen_cost]
        expansion_with_chosen_cost = [expansion for (_, child_cost, expansion)
                                      in children_alternatives_with_cost
                                      if child_cost == chosen_cost]

        index = self.choose_node_expansion(node, children_with_chosen_cost)

        chosen_children = children_with_chosen_cost[index]
        chosen_expansion = expansion_with_chosen_cost[index]
        chosen_children = self.process_chosen_children(
            chosen_children, chosen_expansion)

        return (symbol, chosen_children)

    def symbol_cost(self, symbol, seen = set()):
        expansions = self.grammar[symbol]
        return min(self.expansion_cost(e, seen | {symbol}) for e in expansions)

    def expansion_cost(self, expansion, seen = set()):
        symbols = nonterminals(expansion)

        if len(symbols) == 0:
            return 1

        if (any(s in seen for s in symbols)):
            return float("inf")

        return sum(self.symbol_cost(s, seen) for s in symbols) + 1

    def log_tree(self, tree):
        if self.log:
            print("Tree:", all_terminals(tree))

    def expansion_probabilities(self, expansions, nonterminal="<symbol>"):
        probabilities = [self.opt_prob(expansion) for expansion in expansions]
        prob_dist = self.prob_distribution(probabilities, nonterminal)

        prob_mapping = {}
        for i in range(len(expansions)):
            expansion = expansion_string(expansions[i])
            prob_mapping[expansion] = prob_dist[i]

        return prob_mapping

    def prob_distribution(self, probabilities, nonterminal="<symbol>"):
        epsilon = 0.00001

        n_unspec_probs = probabilities.count(None)
        if n_unspec_probs == 0:
            sum_probs = cast(float, sum(probabilities))
            if abs(sum_probs - 1.0) >= epsilon:
                raise GrammarError(nonterminal + \
                                   ": sum of probabilities must be 1.0")
            return probabilities

        sum_spec_probs = 0.0
        for p in probabilities:
            if p is not None:
                sum_spec_probs += p

        if sum_spec_probs < 0 or sum_spec_probs > 1.0:
           raise GrammarError(nonterminal + \
                              ": sum of probabilities must be in [0.0, 1.0]")

        default_prob = (1.0 - sum_spec_probs) / n_unspec_probs

        all_probs = []
        for p in probabilities:
            if p is None:
                p = default_prob
            all_probs.append(p)

        if abs(sum(all_probs) - 1.0) >= epsilon:
            raise GrammarError(nonterminal + \
                               ": couldn't set default probbilities")

        return all_probs
